@@ -1,27 +1,18 @@
-def warn(*args, **kwargs):
-    pass
-
-
-import warnings
-
-warnings.warn = warn
-
 from abc import abstractmethod, ABC
+from numbers import Number
 from typing import Tuple, Optional, Callable, Any, List, Dict, Union
 from scipy.stats import PearsonRConstantInputWarning
 from sklearn.preprocessing import StandardScaler
-import sklearn.gaussian_process as gp
 import numpy as np
 import scipy as sp
 import time
 from pathlib import Path
 import river as rv
-from joblib import Parallel, delayed
+from collections import deque
 
 from movado.estimator import Estimator
 from movado.mab_handler import MabHandler
 from movado.chained_estimator import ChainedEstimator
-import asyncio
 
 is_call_exact: List[bool] = []  # i-th element True if i-th call is exact
 
@@ -37,9 +28,9 @@ class Controller(ABC):
         **kwargs
     ):
         self.__scaler = StandardScaler()
-        self.__time_dataset: List[float] = []
-        self.__error_dataset: List[float] = []
-        self.__cost_dataset: List[float] = []
+        self.__time_dataset: "deque" = deque()
+        self.__error_dataset: "deque" = deque()
+        self.__cost_dataset: "deque" = deque()
         self.__to_reshape = False
         self._exact: Callable[[List[float]], List[float]] = exact_fitness
         self._estimator: Estimator = estimator
@@ -71,8 +62,13 @@ class Controller(ABC):
             eta=10000,
             verbose=True,
         )
-        self.__fitted_time = []
-        self.__fitted_error = []
+        self.__fitted_time: "deque" = deque()
+        self.__fitted_error: "deque" = deque()
+        self.__fitted_time_mean: float = 0
+        self.__fitted_error_mean: float = 0
+        self.__cost_mean: float = 0
+        self.__time_importance: float = 0
+        self.__error_importance: float = 0
         if self._debug:
             Path("movado_debug").mkdir(exist_ok=True)
             self._controller_debug = "movado_debug/controller.csv"
@@ -183,7 +179,8 @@ class Controller(ABC):
         #    mab_weight_forced_probability=mab_weight_forced_probability,
         #    is_point_in_context=is_point_in_context,
         # )
-
+        if np.isscalar(exact):
+            exact = [exact]
         self._estimator.train(point, exact)
         return exact, exec_time
 
@@ -340,6 +337,14 @@ class Controller(ABC):
         except ValueError:
             return 0
 
+    @staticmethod
+    def __update_mean(old_mean: float, old_sample_number: int, new_sample: float):
+        """
+        Return updated mean in light of new observation without
+        full recalculation.
+        """
+        return (old_sample_number * old_mean + new_sample) / (old_sample_number + 1)
+
     def compute_first_order_sensitivity_index(self):
         self.__time_model.learn_one(
             ChainedEstimator.X_to_river([self.__time_dataset[-1]]),
@@ -350,38 +355,37 @@ class Controller(ABC):
             self.__cost_dataset[-1],
         )
         self.__fitted_time.append(
-            float(
-                self.__time_model.predict_one(
-                    ChainedEstimator.X_to_river([self.__time_dataset[-1]])
-                )
+            self.__time_model.predict_one(
+                ChainedEstimator.X_to_river([self.__time_dataset[-1]])
             )
         )
         self.__fitted_error.append(
-            float(
-                self.__error_model.predict_one(
-                    ChainedEstimator.X_to_river([self.__error_dataset[-1]])
-                )
+            self.__error_model.predict_one(
+                ChainedEstimator.X_to_river([self.__error_dataset[-1]])
             )
         )
         if len(self.__cost_dataset) == 1:
             return 0
-        time_importance = np.sum(
-            [
-                ((fitted - np.mean(self.__fitted_time)) ** 2)
-                / (cost - np.mean(self.__cost_dataset)) ** 2
-                for fitted, cost in zip(self.__fitted_time, self.__cost_dataset)
-            ]
+        self.__fitted_time_mean = self.__update_mean(
+            self.__fitted_time_mean, len(self.__fitted_time) - 1, self.__fitted_time[-1]
         )
-        error_importance = np.sum(
-            [
-                ((fitted - np.mean(self.__fitted_error)) ** 2)
-                / (cost - np.mean(self.__cost_dataset)) ** 2
-                for fitted, cost in zip(self.__fitted_error, self.__cost_dataset)
-            ]
+        self.__fitted_error_mean = self.__update_mean(
+            self.__fitted_error_mean,
+            len(self.__fitted_error) - 1,
+            self.__fitted_error[-1],
         )
-        importance_sum = np.sum([time_importance, error_importance])
-        time_importance = time_importance / importance_sum
-        error_importance = error_importance / importance_sum
+        self.__cost_mean = self.__update_mean(
+            self.__cost_mean, len(self.__cost_dataset) - 1, self.__cost_dataset[-1]
+        )
+        self.__time_importance += (
+            self.__fitted_time[-1] - self.__fitted_time_mean
+        ) ** 2 / (self.__cost_dataset[-1] - self.__cost_mean) ** 2
+        self.__error_importance += (
+            self.__fitted_error[-1] - self.__fitted_error_mean
+        ) ** 2 / (self.__cost_dataset[-1] - self.__cost_mean) ** 2
+        importance_sum = self.__time_importance + self.__error_importance
+        time_importance = self.__time_importance / importance_sum
+        error_importance = self.__error_importance / importance_sum
         return np.abs(time_importance - error_importance)
 
     @staticmethod
