@@ -1,5 +1,7 @@
 from collections import OrderedDict
-from typing import List, Callable
+from numbers import Number
+from typing import List, Callable, Union
+import numpy as np
 
 from movado.controller import Controller
 from movado.estimator import Estimator
@@ -17,6 +19,9 @@ from movado.mab_controller import MabController  # pylint: disable=unused-import
 # noinspection PyUnresolvedReferences
 from movado.voting_controller import VotingController  # pylint: disable=unused-import
 
+# noinspection PyUnresolvedReferences
+from movado.shadow_controller import ShadowController
+
 # noinspection PyUnresolvedReferences
 from movado.kernel_regression_model import (
     KernelRegressionModel,
@@ -34,14 +39,16 @@ def approximate(
         controller: Controller
         estimator: Estimator
         is_first_call: bool = True
+        outputs: int = -1
 
         @wraps(func)
-        def wrapper(*wrapper_args) -> List[float]:
+        def wrapper(*wrapper_args) -> Union[List[float], float]:
             # args must either contain the point to be analyzed
             # or it must contain self (for class methods) and then the point
             nonlocal is_first_call
             nonlocal controller
             nonlocal estimator
+            nonlocal outputs
             if len(wrapper_args) < 1 or len(wrapper_args) > 2:
                 raise Exception(
                     "The decorated function must have a single input which is a list of numbers, "
@@ -59,7 +66,10 @@ def approximate(
                     HoeffdingAdaptiveTreeModel if not selected_model else selected_model
                 )
                 # TODO add multi-output non-chained estimators
-                outputs: int = kwargs.get("outputs")
+                outputs = kwargs.get("outputs")
+                voters: int = kwargs.get("voters")
+                if not voters:
+                    voters = 750
                 if not outputs:
                     raise Exception(
                         "Please specify outputs as a kwarg as the number of targets of the fitness function"
@@ -85,38 +95,51 @@ def approximate(
                 if selected_controller is MabController:
                     params = OrderedDict(
                         {
-                            "cover": [3, 5, 7, 11, 20],
-                            "mab_weight_epsilon": [0.1, 0.2, 0.3, 0.4],
-                            "mab_weight_bandwidth": [1, 5, 10, 20, 50, 75],
+                            "cover": list(range(3, 20)),
+                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
+                            "mab_weight_bandwidth": list(range(3, 7)),
                         }
                     )
                 else:
-                    # All other cases are handled as a DistanceController
                     params = OrderedDict(
                         {
-                            "nth_nearest": [1, 3, 5, 7],
-                            "mab_epsilon": [0.1, 0.2, 0.3],
-                            "mab_bandwidth": [1, 5, 10, 20],
-                            "mab_weight_epsilon": [0.1, 0.2, 0.3],
-                            "mab_weight_bandwidth": [1, 5, 10, 20],
+                            "nth_nearest": list(range(1, 6)),
+                            "mab_epsilon": [i / 100 for i in range(1, 41)],
+                            "mab_bandwidth": list(range(3, 7)),
+                            "mab_weight_epsilon": [i / 100 for i in range(1, 41)],
+                            "mab_weight_bandwidth": list(range(3, 7)),
                         }
                     )
-                controller = VotingController(
-                    controller,
-                    estimator,
-                    func,
-                    params,
-                    self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
-                    **{
-                        k[k.find("_") + 1 :]: v
-                        for k, v in kwargs.items()
-                        if k[: k.find("_")] == "controller"
-                    },
-                )
-
-            return list(
-                controller.compute_objective(point)
-            )  # we expect the first and only argument of the function to be the input point
+                available_models = int(np.prod([len(v) for v in params.values()]))
+                voters = voters if voters <= available_models else available_models
+                if voters <= 1:
+                    controller = controller(
+                        exact_fitness=func,
+                        estimator=estimator,
+                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
+                        **{
+                            k[k.find("_") + 1 :]: v
+                            for k, v in kwargs.items()
+                            if k[: k.find("_")] == "controller"
+                        },
+                    )
+                else:
+                    controller = VotingController(
+                        controller,
+                        estimator,
+                        func,
+                        voters,
+                        params,
+                        self_exact=None if len(wrapper_args) == 1 else wrapper_args[0],
+                        **{
+                            k[k.find("_") + 1 :]: v
+                            for k, v in kwargs.items()
+                            if k[: k.find("_")] == "controller"
+                        },
+                    )
+            if outputs == 1:
+                return list(controller.compute_objective(point))[0]
+            return list(controller.compute_objective(point))
 
         if kwargs.get("disabled"):
             return func
